@@ -17,6 +17,10 @@ import calendar
 import binascii
 import random
 
+# as pytz is optional in thirdparty libs but we need it for good support under
+# python2, just test that it's well installed
+import pytz  # noqa
+
 try:
     from collections import OrderedDict
 except ImportError:
@@ -44,7 +48,29 @@ re_star = re.compile('[*]')
 hash_expression_re = re.compile(
     r'^(?P<hash_type>h|r)(\((?P<range_begin>\d+)-(?P<range_end>\d+)\))?(\/(?P<divisor>\d+))?$'
 )
-VALID_LEN_EXPRESSION = [5, 6]
+MINUTE_FIELD = 0
+HOUR_FIELD = 1
+DAY_FIELD = 2
+MONTH_FIELD = 3
+DOW_FIELD = 4
+SECOND_FIELD = 5
+YEAR_FIELD = 6
+UNIX_FIELDS = (MINUTE_FIELD, HOUR_FIELD, DAY_FIELD, MONTH_FIELD, DOW_FIELD)
+SECOND_FIELDS = (MINUTE_FIELD, HOUR_FIELD, DAY_FIELD, MONTH_FIELD, DOW_FIELD, SECOND_FIELD)
+YEAR_FIELDS = (MINUTE_FIELD, HOUR_FIELD, DAY_FIELD, MONTH_FIELD, DOW_FIELD, SECOND_FIELD, YEAR_FIELD)
+CRON_FIELDS = {
+    'unix': UNIX_FIELDS,
+    'second': SECOND_FIELDS,
+    'year': YEAR_FIELDS,
+    len(UNIX_FIELDS): UNIX_FIELDS,
+    len(SECOND_FIELDS): SECOND_FIELDS,
+    len(YEAR_FIELDS): YEAR_FIELDS,
+}
+UNIX_CRON_LEN = len(UNIX_FIELDS)
+SECOND_CRON_LEN = len(SECOND_FIELDS)
+YEAR_CRON_LEN = len(YEAR_FIELDS)
+# retrocompat
+VALID_LEN_EXPRESSION = set([a for a in CRON_FIELDS if isinstance(a, int)])
 EXPRESSIONS = {}
 
 
@@ -112,7 +138,8 @@ class croniter(object):
         (1, 31),
         (1, 12),
         (0, 7),
-        (0, 59)
+        (0, 59),
+        (1970, 2099)
     )
     DAYS = (
         31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
@@ -126,7 +153,9 @@ class croniter(object):
         copy.deepcopy(M_ALPHAS),
         # 4: dow
         copy.deepcopy(DOW_ALPHAS),
-        # command/user
+        # 5: second
+        {},
+        # 6: year
         {}
     )
 
@@ -137,6 +166,7 @@ class croniter(object):
         {0: 1},
         {7: 0},
         {},
+        {}
     )
 
     LEN_MEANS_ALL = (
@@ -145,18 +175,22 @@ class croniter(object):
         31,
         12,
         7,
-        60
+        60,
+        130
     )
 
-    bad_length = 'Exactly 5 or 6 columns has to be specified for iterator ' \
+    bad_length = 'Exactly 5, 6 or 7 columns has to be specified for iterator ' \
                  'expression.'
 
     def __init__(self, expr_format, start_time=None, ret_type=float,
                  day_or=True, max_years_between_matches=None, is_prev=False,
-                 hash_id=None, implement_cron_bug=False):
+                 hash_id=None, implement_cron_bug=False, second_at_beginning=None,
+                 expand_from_start_time=False):
         self._ret_type = ret_type
         self._day_or = day_or
         self._implement_cron_bug = implement_cron_bug
+        self.second_at_beginning = bool(second_at_beginning)
+        self._expand_from_start_time = expand_from_start_time
 
         if hash_id:
             if not isinstance(hash_id, (bytes, str)):
@@ -180,8 +214,14 @@ class croniter(object):
         self.cur = None
         self.set_current(start_time, force=False)
 
-        self.expanded, self.nth_weekday_of_month = self.expand(expr_format, hash_id=hash_id)
-        self.expressions = EXPRESSIONS[(expr_format, hash_id)]
+        self.expanded, self.nth_weekday_of_month = self.expand(
+            expr_format,
+            hash_id=hash_id,
+            from_timestamp=self.dst_start_time if self._expand_from_start_time else None,
+            second_at_beginning=second_at_beginning
+        )
+        self.fields = CRON_FIELDS[len(self.expanded)]
+        self.expressions = EXPRESSIONS[(expr_format, hash_id, second_at_beginning)]
         self._is_prev = is_prev
 
     @classmethod
@@ -192,12 +232,19 @@ class croniter(object):
             raise CroniterNotAlphaError(
                 "[{0}] is not acceptable".format(" ".join(expressions)))
 
-    def get_next(self, ret_type=None, start_time=None):
-        self.set_current(start_time, force=True)
-        return self._get_next(ret_type or self._ret_type, is_prev=False)
+    def get_next(self, ret_type=None, start_time=None, update_current=True):
+        if start_time and self._expand_from_start_time:
+            raise ValueError("start_time is not supported when using expand_from_start_time = True.")
+        return self._get_next(ret_type or self._ret_type,
+                              start_time=start_time,
+                              is_prev=False,
+                              update_current=update_current)
 
-    def get_prev(self, ret_type=None):
-        return self._get_next(ret_type or self._ret_type, is_prev=True)
+    def get_prev(self, ret_type=None, start_time=None, update_current=True):
+        return self._get_next(ret_type or self._ret_type,
+                              start_time=start_time,
+                              is_prev=True,
+                              update_current=update_current)
 
     def get_current(self, ret_type=None):
         ret_type = ret_type or self._ret_type
@@ -227,7 +274,7 @@ class croniter(object):
         """
         Converts a UNIX timestamp `timestamp` into a `datetime` object.
         """
-        result = datetime.datetime.utcfromtimestamp(timestamp)
+        result = datetime.datetime.fromtimestamp(timestamp, tz=tzutc()).replace(tzinfo=None)
         if self.tzinfo:
             result = result.replace(tzinfo=tzutc()).astimezone(self.tzinfo)
 
@@ -243,7 +290,9 @@ class croniter(object):
         """
         return timedelta_to_seconds(td)
 
-    def _get_next(self, ret_type=None, start_time=None, is_prev=None):
+    def _get_next(self, ret_type=None, start_time=None, is_prev=None, update_current=None):
+        if update_current is None:
+            update_current = True
         self.set_current(start_time, force=True)
         if is_prev is None:
             is_prev = self._is_prev
@@ -259,20 +308,23 @@ class croniter(object):
 
         # exception to support day of month and day of week as defined in cron
         dom_dow_exception_processed = False
-        if (expanded[2][0] != '*' and expanded[4][0] != '*') and self._day_or:
+        if (expanded[DAY_FIELD][0] != '*' and expanded[DOW_FIELD][0] != '*') and self._day_or:
             # If requested, handle a bug in vixie cron/ISC cron where day_of_month and day_of_week form
             # an intersection (AND) instead of a union (OR) if either field is an asterisk or starts with an asterisk
             # (https://crontab.guru/cron-bug.html)
-            if self._implement_cron_bug and (re_star.match(self.expressions[2]) or re_star.match(self.expressions[4])):
+            if (
+                self._implement_cron_bug and
+                (re_star.match(self.expressions[DAY_FIELD]) or re_star.match(self.expressions[DOW_FIELD]))
+            ):
                 # To produce a schedule identical to the cron bug, we'll bypass the code that
                 # makes a union of DOM and DOW, and instead skip to the code that does an intersect instead
                 pass
             else:
-                bak = expanded[4]
-                expanded[4] = ['*']
+                bak = expanded[DOW_FIELD]
+                expanded[DOW_FIELD] = ['*']
                 t1 = self._calc(self.cur, expanded, nth_weekday_of_month, is_prev)
-                expanded[4] = bak
-                expanded[2] = ['*']
+                expanded[DOW_FIELD] = bak
+                expanded[DAY_FIELD] = ['*']
 
                 t2 = self._calc(self.cur, expanded, nth_weekday_of_month, is_prev)
                 if not is_prev:
@@ -315,7 +367,8 @@ class croniter(object):
                     dtresult = dtresult_adjusted
                     result = result_adjusted
                 self.dst_start_time = result
-        self.cur = result
+        if update_current:
+            self.cur = result
         if issubclass(ret_type, datetime.datetime):
             result = dtresult
         return result
@@ -324,7 +377,7 @@ class croniter(object):
     # objects in a loop, like "for dt in croniter('5 0 * * *'): ..."
     # or for combining multiple croniters into single
     # dates feed using 'itertools' module
-    def all_next(self, ret_type=None):
+    def all_next(self, ret_type=None, start_time=None, update_current=None):
         '''Generator of all consecutive dates. Can be used instead of
         implicit call to __iter__, whenever non-default
         'ret_type' has to be specified.
@@ -333,19 +386,23 @@ class croniter(object):
         try:
             while True:
                 self._is_prev = False
-                yield self._get_next(ret_type or self._ret_type)
+                yield self._get_next(ret_type or self._ret_type,
+                                     start_time=start_time, update_current=update_current)
+                start_time = None
         except CroniterBadDateError:
             if self._max_years_btw_matches_explicitly_set:
                 return
             else:
                 raise
 
-    def all_prev(self, ret_type=None):
+    def all_prev(self, ret_type=None, start_time=None, update_current=None):
         '''Generator of all previous dates.'''
         try:
             while True:
                 self._is_prev = True
-                yield self._get_next(ret_type or self._ret_type)
+                yield self._get_next(ret_type or self._ret_type,
+                                     start_time=start_time, update_current=update_current)
+                start_time = None
         except CroniterBadDateError:
             if self._max_years_btw_matches_explicitly_set:
                 return
@@ -364,12 +421,12 @@ class croniter(object):
             now = math.ceil(now)
             nearest_diff_method = self._get_prev_nearest_diff
             sign = -1
-            offset = (len(expanded) == 6 or now % 60 > 0) and 1 or 60
+            offset = (len(expanded) > UNIX_CRON_LEN or now % 60 > 0) and 1 or 60
         else:
             now = math.floor(now)
             nearest_diff_method = self._get_next_nearest_diff
             sign = 1
-            offset = (len(expanded) == 6) and 1 or 60
+            offset = (len(expanded) > UNIX_CRON_LEN) and 1 or 60
 
         dst = now = self._timestamp_to_datetime(now + sign * offset)
 
@@ -377,12 +434,31 @@ class croniter(object):
         current_year = now.year
         DAYS = self.DAYS
 
+        def proc_year(d):
+            if len(expanded) == YEAR_CRON_LEN:
+                try:
+                    expanded[YEAR_FIELD].index("*")
+                except ValueError:
+                    # use None as range_val to indicate no loop
+                    diff_year = nearest_diff_method(d.year, expanded[YEAR_FIELD], None)
+                    if diff_year is None:
+                        return None, d
+                    elif diff_year != 0:
+                        if is_prev:
+                            d += relativedelta(years=diff_year, month=12, day=31,
+                                               hour=23, minute=59, second=59)
+                        else:
+                            d += relativedelta(years=diff_year, month=1, day=1,
+                                               hour=0, minute=0, second=0)
+                        return True, d
+            return False, d
+
         def proc_month(d):
             try:
-                expanded[3].index('*')
+                expanded[MONTH_FIELD].index('*')
             except ValueError:
                 diff_month = nearest_diff_method(
-                    d.month, expanded[3], self.MONTHS_IN_YEAR)
+                    d.month, expanded[MONTH_FIELD], self.MONTHS_IN_YEAR)
                 days = DAYS[month - 1]
                 if month == 2 and self.is_leap(year) is True:
                     days += 1
@@ -393,6 +469,8 @@ class croniter(object):
                     if is_prev:
                         d += relativedelta(months=diff_month)
                         reset_day = DAYS[d.month - 1]
+                        if d.month == 2 and self.is_leap(d.year) is True:
+                            reset_day += 1
                         d += relativedelta(
                             day=reset_day, hour=23, minute=59, second=59)
                     else:
@@ -403,21 +481,21 @@ class croniter(object):
 
         def proc_day_of_month(d):
             try:
-                expanded[2].index('*')
+                expanded[DAY_FIELD].index('*')
             except ValueError:
                 days = DAYS[month - 1]
                 if month == 2 and self.is_leap(year) is True:
                     days += 1
-                if 'l' in expanded[2] and days == d.day:
+                if 'l' in expanded[DAY_FIELD] and days == d.day:
                     return False, d
 
                 if is_prev:
                     days_in_prev_month = DAYS[
                         (month - 2) % self.MONTHS_IN_YEAR]
                     diff_day = nearest_diff_method(
-                        d.day, expanded[2], days_in_prev_month)
+                        d.day, expanded[DAY_FIELD], days_in_prev_month)
                 else:
-                    diff_day = nearest_diff_method(d.day, expanded[2], days)
+                    diff_day = nearest_diff_method(d.day, expanded[DAY_FIELD], days)
 
                 if diff_day is not None and diff_day != 0:
                     if is_prev:
@@ -431,10 +509,10 @@ class croniter(object):
 
         def proc_day_of_week(d):
             try:
-                expanded[4].index('*')
+                expanded[DOW_FIELD].index('*')
             except ValueError:
                 diff_day_of_week = nearest_diff_method(
-                    d.isoweekday() % 7, expanded[4], 7)
+                    d.isoweekday() % 7, expanded[DOW_FIELD], 7)
                 if diff_day_of_week is not None and diff_day_of_week != 0:
                     if is_prev:
                         d += relativedelta(days=diff_day_of_week,
@@ -497,9 +575,9 @@ class croniter(object):
 
         def proc_hour(d):
             try:
-                expanded[1].index('*')
+                expanded[HOUR_FIELD].index('*')
             except ValueError:
-                diff_hour = nearest_diff_method(d.hour, expanded[1], 24)
+                diff_hour = nearest_diff_method(d.hour, expanded[HOUR_FIELD], 24)
                 if diff_hour is not None and diff_hour != 0:
                     if is_prev:
                         d += relativedelta(
@@ -511,9 +589,9 @@ class croniter(object):
 
         def proc_minute(d):
             try:
-                expanded[0].index('*')
+                expanded[MINUTE_FIELD].index('*')
             except ValueError:
-                diff_min = nearest_diff_method(d.minute, expanded[0], 60)
+                diff_min = nearest_diff_method(d.minute, expanded[MINUTE_FIELD], 60)
                 if diff_min is not None and diff_min != 0:
                     if is_prev:
                         d += relativedelta(minutes=diff_min, second=59)
@@ -523,11 +601,11 @@ class croniter(object):
             return False, d
 
         def proc_second(d):
-            if len(expanded) == 6:
+            if len(expanded) > UNIX_CRON_LEN:
                 try:
-                    expanded[5].index('*')
+                    expanded[SECOND_FIELD].index('*')
                 except ValueError:
-                    diff_sec = nearest_diff_method(d.second, expanded[5], 60)
+                    diff_sec = nearest_diff_method(d.second, expanded[SECOND_FIELD], 60)
                     if diff_sec is not None and diff_sec != 0:
                         d += relativedelta(seconds=diff_sec)
                         return True, d
@@ -535,7 +613,8 @@ class croniter(object):
                 d += relativedelta(second=0)
             return False, d
 
-        procs = [proc_month,
+        procs = [proc_year,
+                 proc_month,
                  proc_day_of_month,
                  (proc_day_of_week_nth if nth_weekday_of_month
                      else proc_day_of_week),
@@ -545,12 +624,20 @@ class croniter(object):
 
         while abs(year - current_year) <= self._max_years_between_matches:
             next = False
+            stop = False
             for proc in procs:
                 (changed, dst) = proc(dst)
+                # `None` can be set mostly for year processing
+                # so please see proc_year / _get_prev_nearest_diff / _get_next_nearest_diff
+                if changed is None:
+                    stop = True
+                    break
                 if changed:
                     month, year = dst.month, dst.year
                     next = True
                     break
+            if stop:
+                break
             if next:
                 continue
             return self._datetime_to_timestamp(dst.replace(microsecond=0))
@@ -574,16 +661,32 @@ class croniter(object):
         return small[0]
 
     def _get_next_nearest_diff(self, x, to_check, range_val):
+        """
+        `range_val` is the range of a field.
+        If no available time, we can move to next loop(like next month).
+        `range_val` can also be set to `None` to indicate that there is no loop.
+        ( Currently, should only used for `year` field )
+        """
         for i, d in enumerate(to_check):
-            if d == "l":
+            if d == "l" and range_val is not None:
                 # if 'l' then it is the last day of month
                 # => its value of range_val
                 d = range_val
             if d >= x:
                 return d - x
+        # When range_val is None and x not exists in to_check,
+        # `None` will be returned to suggest no more available time
+        if range_val is None:
+            return None
         return to_check[0] - x + range_val
 
     def _get_prev_nearest_diff(self, x, to_check, range_val):
+        """
+        `range_val` is the range of a field.
+        If no available time, we can move to previous loop(like previous month).
+        Range_val can also be set to `None` to indicate that there is no loop.
+        ( Currently should only used for `year` field )
+        """
         candidates = to_check[:]
         candidates.reverse()
         for d in candidates:
@@ -591,6 +694,10 @@ class croniter(object):
                 return d - x
         if 'l' in candidates:
             return -x
+        # When range_val is None and x not exists in to_check,
+        # `None` will be returned to suggest no more available time
+        if range_val is None:
+            return None
         candidate = candidates[0]
         for c in candidates:
             # fixed: c < range_val
@@ -626,7 +733,7 @@ class croniter(object):
             return False
 
     @classmethod
-    def _expand(cls, expr_format, hash_id=None):
+    def _expand(cls, expr_format, hash_id=None, second_at_beginning=False, from_timestamp=None):
         # Split the expression in components, and normalize L -> l, MON -> mon,
         # etc. Keep expr_format untouched so we can use it in the exception
         # messages.
@@ -652,12 +759,28 @@ class croniter(object):
         if len(expressions) not in VALID_LEN_EXPRESSION:
             raise CroniterBadCronError(cls.bad_length)
 
+        if len(expressions) > UNIX_CRON_LEN and second_at_beginning:
+            # move second to it's own(6th) field to process by same logical
+            expressions.insert(SECOND_FIELD, expressions.pop(0))
+
         expanded = []
         nth_weekday_of_month = {}
 
         for i, expr in enumerate(expressions):
             for expanderid, expander in EXPANDERS.items():
-                expr = expander(cls).expand(efl, i, expr, hash_id=hash_id)
+                expr = expander(cls).expand(efl, i, expr, hash_id=hash_id, from_timestamp=from_timestamp)
+
+            if "?" in expr:
+                if expr != "?":
+                    raise CroniterBadCronError(
+                        "[{0}] is not acceptable.  Question mark can not "
+                        "used with other characters".format(expr_format))
+                if i not in [DAY_FIELD, DOW_FIELD]:
+                    raise CroniterBadCronError(
+                        "[{0}] is not acceptable.  Question mark can only used "
+                        "in day_of_month or day_of_week".format(expr_format))
+                # currently just trade `?` as `*`
+                expr = "*"
 
             e_list = expr.split(',')
             res = []
@@ -666,7 +789,7 @@ class croniter(object):
                 e = e_list.pop()
                 nth = None
 
-                if i == 4:
+                if i == DOW_FIELD:
                     # Handle special case in the dow expression: 2#3, l3
                     special_dow_rem = special_dow_re.match(str(e))
                     if special_dow_rem:
@@ -706,7 +829,7 @@ class croniter(object):
                     # early abort if low/high are out of bounds
 
                     (low, high, step) = m.group(1), m.group(2), m.group(4) or 1
-                    if i == 2 and high == 'l':
+                    if i == DAY_FIELD and high == 'l':
                         high = '31'
 
                     if not only_int_re.search(low):
@@ -719,7 +842,7 @@ class croniter(object):
                         not low or not high or int(low) > int(high)
                         or not only_int_re.search(str(step))
                     ):
-                        if i == 4 and high == '0':
+                        if i == DOW_FIELD and high == '0':
                             # handle -Sun notation -> 7
                             high = '7'
                         else:
@@ -732,13 +855,17 @@ class croniter(object):
                     ):
                         raise CroniterBadCronError(
                             "{0} is out of bands".format(expr_format))
+
+                    if from_timestamp:
+                        low = cls._get_low_from_current_date_number(i, int(step), int(from_timestamp))
+
                     try:
                         rng = range(low, high + 1, step)
                     except ValueError as exc:
                         raise CroniterBadCronError(
                             'invalid range: {0}'.format(exc))
                     e_list += (["{0}#{1}".format(item, nth) for item in rng]
-                               if i == 4 and nth and nth != "l" else rng)
+                               if i == DOW_FIELD and nth and nth != "l" else rng)
                 else:
                     if t.startswith('-'):
                         raise CroniterBadCronError((
@@ -754,11 +881,12 @@ class croniter(object):
                         pass
 
                     if t in cls.LOWMAP[i] and not (
-                        # do not support 0 as a month either for classical 5 fields cron
-                        # or 6fields second repeat form
+                        # do not support 0 as a month either for classical 5 fields cron,
+                        # 6fields second repeat form or 7 fields year form
                         # but still let conversion happen if day field is shifted
-                        (i in [2, 3] and len(expressions) == 5) or
-                        (i in [3, 4] and len(expressions) == 6)
+                        (i in [DAY_FIELD, MONTH_FIELD] and len(expressions) == UNIX_CRON_LEN) or
+                        (i in [MONTH_FIELD, DOW_FIELD] and len(expressions) == SECOND_CRON_LEN) or
+                        (i in [DAY_FIELD, MONTH_FIELD, DOW_FIELD] and len(expressions) == YEAR_CRON_LEN)
                     ):
                         t = cls.LOWMAP[i][t]
 
@@ -773,7 +901,7 @@ class croniter(object):
 
                     res.append(t)
 
-                    if i == 4 and nth:
+                    if i == DOW_FIELD and nth:
                         if t not in nth_weekday_of_month:
                             nth_weekday_of_month[t] = set()
                         nth_weekday_of_month[t].add(nth)
@@ -781,7 +909,14 @@ class croniter(object):
             res = set(res)
             res = sorted(res, key=lambda i: "{:02}".format(i) if isinstance(i, int) else i)
             if len(res) == cls.LEN_MEANS_ALL[i]:
-                res = ['*']
+                # Make sure the wildcard is used in the correct way (avoid over-optimization)
+                if (
+                    (i == DAY_FIELD and '*' not in expressions[DOW_FIELD]) or
+                    (i == DOW_FIELD and '*' not in expressions[DAY_FIELD])
+                ):
+                    pass
+                else:
+                    res = ['*']
 
             expanded.append(['*'] if (len(res) == 1
                                       and res[0] == '*')
@@ -789,22 +924,25 @@ class croniter(object):
 
         # Check to make sure the dow combo in use is supported
         if nth_weekday_of_month:
-            dow_expanded_set = set(expanded[4])
+            dow_expanded_set = set(expanded[DOW_FIELD])
             dow_expanded_set = dow_expanded_set.difference(nth_weekday_of_month.keys())
             dow_expanded_set.discard("*")
-            if dow_expanded_set:
+            # Skip: if it's all weeks instead of wildcard
+            if dow_expanded_set and len(set(expanded[DOW_FIELD])) != cls.LEN_MEANS_ALL[DOW_FIELD]:
                 raise CroniterUnsupportedSyntaxError(
                     "day-of-week field does not support mixing literal values and nth day of week syntax.  "
                     "Cron: '{}'    dow={} vs nth={}".format(expr_format, dow_expanded_set, nth_weekday_of_month))
 
-        EXPRESSIONS[(expr_format, hash_id)] = expressions
+        EXPRESSIONS[(expr_format, hash_id, second_at_beginning)] = expressions
         return expanded, nth_weekday_of_month
 
     @classmethod
-    def expand(cls, expr_format, hash_id=None):
+    def expand(cls, expr_format, hash_id=None, second_at_beginning=False, from_timestamp=None):
         """Shallow non Croniter ValueError inside a nice CroniterBadCronError"""
         try:
-            return cls._expand(expr_format, hash_id=hash_id)
+            return cls._expand(expr_format, hash_id=hash_id,
+                               second_at_beginning=second_at_beginning,
+                               from_timestamp=from_timestamp)
         except (ValueError,) as exc:
             error_type, error_instance, traceback = sys.exc_info()
             if isinstance(exc, CroniterError):
@@ -817,28 +955,63 @@ class croniter(object):
                 raise CroniterBadCronError("{0}".format(exc))
 
     @classmethod
-    def is_valid(cls, expression, hash_id=None):
+    def _get_low_from_current_date_number(cls, i, step, from_timestamp):
+        dt = datetime.datetime.fromtimestamp(from_timestamp, tz=datetime.timezone.utc)
+        if i == MINUTE_FIELD:
+            return dt.minute % step
+        if i == HOUR_FIELD:
+            return dt.hour % step
+        if i == DAY_FIELD:
+            return ((dt.day - 1) % step) + 1
+        if i == MONTH_FIELD:
+            return dt.month % step
+        if i == DOW_FIELD:
+            return (dt.weekday() + 1) % step
+
+        raise ValueError("Can't get current date number for index larger than 4")
+
+    @classmethod
+    def is_valid(cls, expression, hash_id=None, encoding='UTF-8',
+                 second_at_beginning=False):
+        if hash_id:
+            if not isinstance(hash_id, (bytes, str)):
+                raise TypeError('hash_id must be bytes or UTF-8 string')
+            if not isinstance(hash_id, bytes):
+                hash_id = hash_id.encode(encoding)
         try:
-            cls.expand(expression, hash_id=hash_id)
+            cls.expand(expression, hash_id=hash_id,
+                       second_at_beginning=second_at_beginning)
         except CroniterError:
             return False
         else:
             return True
 
     @classmethod
-    def match(cls, cron_expression, testdate, day_or=True):
-        cron = cls(cron_expression, testdate, ret_type=datetime.datetime, day_or=day_or)
-        td, ms1 = cron.get_current(datetime.datetime), relativedelta(microseconds=1)
-        if not td.microsecond:
-            td = td + ms1
-        cron.set_current(td, force=True)
-        tdp, tdt = cron.get_current(), cron.get_prev()
-        precision_in_seconds = 1 if len(cron.expanded) == 6 else 60
-        return (max(tdp, tdt) - min(tdp, tdt)).total_seconds() < precision_in_seconds
+    def match(cls, cron_expression, testdate, day_or=True, second_at_beginning=False):
+        return cls.match_range(cron_expression, testdate, testdate, day_or, second_at_beginning)
+
+    @classmethod
+    def match_range(cls, cron_expression, from_datetime, to_datetime,
+                    day_or=True, second_at_beginning=False):
+        cron = cls(cron_expression, to_datetime, ret_type=datetime.datetime,
+                   day_or=day_or, second_at_beginning=second_at_beginning)
+        tdp = cron.get_current(datetime.datetime)
+        if not tdp.microsecond:
+            tdp += relativedelta(microseconds=1)
+        cron.set_current(tdp, force=True)
+        try:
+            tdt = cron.get_prev()
+        except CroniterBadDateError:
+            return False
+        precision_in_seconds = 1 if len(cron.expanded) > UNIX_CRON_LEN else 60
+        duration_in_second = (to_datetime - from_datetime).total_seconds() + precision_in_seconds
+        return (max(tdp, tdt) - min(tdp, tdt)).total_seconds() < duration_in_second
 
 
 def croniter_range(start, stop, expr_format, ret_type=None, day_or=True, exclude_ends=False,
-                   _croniter=None):
+                   _croniter=None,
+                   second_at_beginning=False,
+                   expand_from_start_time=False):
     """
     Generator that provides all times from start to stop matching the given cron expression.
     If the cron expression matches either 'start' and/or 'stop', those times will be returned as
@@ -851,7 +1024,7 @@ def croniter_range(start, stop, expr_format, ret_type=None, day_or=True, exclude
     auto_rt = datetime.datetime
     # type is used in first if branch for perfs reasons
     if (
-        type(start) != type(stop) and not (
+        type(start) is not type(stop) and not (
             isinstance(start, type(stop)) or
             isinstance(stop, type(start)))
     ):
@@ -859,7 +1032,7 @@ def croniter_range(start, stop, expr_format, ret_type=None, day_or=True, exclude
             "The start and stop must be same type.  {0} != {1}".
             format(type(start), type(stop)))
     if isinstance(start, (float, int)):
-        start, stop = (datetime.datetime.utcfromtimestamp(t) for t in (start, stop))
+        start, stop = (datetime.datetime.fromtimestamp(t, tzutc()).replace(tzinfo=None) for t in (start, stop))
         auto_rt = float
     if ret_type is None:
         ret_type = auto_rt
@@ -873,7 +1046,9 @@ def croniter_range(start, stop, expr_format, ret_type=None, day_or=True, exclude
             stop -= ms1
     year_span = math.floor(abs(stop.year - start.year)) + 1
     ic = _croniter(expr_format, start, ret_type=datetime.datetime, day_or=day_or,
-                   max_years_between_matches=year_span)
+                   max_years_between_matches=year_span,
+                   second_at_beginning=second_at_beginning,
+                   expand_from_start_time=expand_from_start_time)
     # define a continue (cont) condition function and step function for the main while loop
     if start < stop:        # Forward
         def cont(v):
@@ -903,20 +1078,15 @@ class HashExpander:
 
     def do(self, idx, hash_type="h", hash_id=None, range_end=None, range_begin=None):
         """Return a hashed/random integer given range/hash information"""
-        hours_or_minutes = idx in {0, 1}
         if range_end is None:
             range_end = self.cron.RANGES[idx][1]
-            if hours_or_minutes:
-                range_end += 1
         if range_begin is None:
             range_begin = self.cron.RANGES[idx][0]
         if hash_type == 'r':
             crc = random.randint(0, 0xFFFFFFFF)
         else:
             crc = binascii.crc32(hash_id) & 0xFFFFFFFF
-        if not hours_or_minutes:
-            return ((crc >> idx) % (range_end - range_begin + 1)) + range_begin
-        return ((crc >> idx) % (range_end - range_begin)) + range_begin
+        return ((crc >> idx) % (range_end - range_begin + 1)) + range_begin
 
     def match(self, efl, idx, expr, hash_id=None, **kw):
         return hash_expression_re.match(expr)
@@ -946,8 +1116,9 @@ class HashExpander:
                     idx,
                     hash_type=m['hash_type'],
                     hash_id=hash_id,
-                    range_end=int(m['divisor']),
-                ) + int(m['range_begin']),
+                    range_begin=int(m['range_begin']),
+                    range_end=int(m['divisor']) - 1 + int(m['range_begin']),
+                ),
                 int(m['range_end']),
                 int(m['divisor']),
             )
@@ -972,7 +1143,8 @@ class HashExpander:
                     idx,
                     hash_type=m['hash_type'],
                     hash_id=hash_id,
-                    range_end=int(m['divisor']),
+                    range_begin=self.cron.RANGES[idx][0],
+                    range_end=int(m['divisor']) - 1 + self.cron.RANGES[idx][0],
                 ),
                 self.cron.RANGES[idx][1],
                 int(m['divisor']),
